module ArgParser
       ( Options(..)
       , Command(..)
       , InitOptions(..)
       , CreateOptions(..)
       , RemoveOptions(..)
       , parseOptions)
       where

import Options.Applicative
import Data.List.Split (splitOn)
import Data.Maybe (fromJust)
import Data.List (find, isPrefixOf)
import System.Directory

import Types (QvmNetwork(..), QvmDisk(..))

-- qvm init
-- qvm create vm1 --core 2 --ram 512M --network predefined:default --disk /path/to/foo.img

withInfo :: Parser a -> String -> ParserInfo a
withInfo opts desc = info (helper <*> opts) $ progDesc desc

data Options =
  Options {optCommand :: Command}
  deriving (Show)

data Command
  = Init InitOptions
  | Create CreateOptions
  | Remove RemoveOptions
  deriving (Show)

data InitOptions =
  InitOptions {noop :: Bool}
  deriving (Show)

data CreateOptions =
  CreateOptions {optName :: String
                ,optCore :: Int
                ,optRamMB :: Int
                ,optNetwork :: QvmNetwork
                ,optDisk :: QvmDisk
                ,optIPAddress :: Maybe String
                ,optGateway :: Maybe String
                ,optDNS :: Maybe String
                ,optNoStart :: Bool}
  deriving (Show)

data RemoveOptions =
  RemoveOptions {removeOptName :: String}
  deriving (Show)

parseInit :: Parser Command
parseInit = Init <$> (InitOptions
    <$> switch (long "noop" <> help "this option does nothing"
                <> hidden))

-- | parse network spec. supported spec is
-- predefined:default
-- predefined:ovs1
-- ovsbridge:name=br0
-- ovsbridge:name=br0,vlan=200
parseNetworkSpec :: String -> ReadM QvmNetwork
parseNetworkSpec s =
  case splitOn ":" s of
    type_:args:[] -> if type_ == "predefined" then
                         return $ DefinedNetwork args
                     else
                         if type_ == "ovsbridge" then
                             let alist = map (splitOn "=") (splitOn "," args)
                                 afind k = find (\l -> head l == k) alist
                                 bridgeName = fromJust (afind "name") !! 1 in
                             case afind "vlan" of
                               Nothing -> return $ OvsBridge bridgeName
                               Just [_, vlanId] -> return $ OvsBridgeWithVLAN bridgeName (read vlanId :: Int)
                         else
                             readerError $ "unsupported network spec: " ++ s
    _ -> readerError $ "unsupported network spec: " ++ s

-- | parse disk spec. supported spec is
-- http://path/to/foo.qcow2
-- http2://path/to/foo.qcow2
-- /path/to/foo.qcow2
parseDiskSpec :: String -> ReadM QvmDisk
parseDiskSpec s =
    if "http://" `isPrefixOf` s || "https://" `isPrefixOf` s then
        return $ HTTPImage s
    else
        return $ LocalImage s

parseCreate :: Parser Command
parseCreate = Create <$> (CreateOptions
    <$> (argument str (metavar "VM_NAME"))
    <*> (option auto (long "core" <> short 'c' <> metavar "N" <> value 2
                      <> help "how may cores for the VM"))
    <*> (option auto (long "ram" <> short 'm' <> metavar "MB" <> value 2048
                      <> help "how may cores for the VM"))
    <*> (option (str >>= parseNetworkSpec)
         (long "network" <> short 'n' <> value DefaultNAT
          <> help "network specification, supported spec is predefined:default or ovsbridge:name=br0 or ovsbridge:name=br0,vlan=200"))
    <*> (option (str >>= parseDiskSpec)
         (long "disk" <> short 'd' <> value (HTTPImage "http://dfs.game.yy.com:8080/download/cirros-0.3.4-x86_64-disk.img")
          <> help "disk image specification, support local qcow2 file and URL to qcow2 image."))
    <*> (optional $ strOption (long "ip" <> metavar "IP"
                               <> help "metadata: IP address"))
    <*> (optional $ strOption (long "gw" <> metavar "GATEWAY"
                               <> help "metadata: gateway address"))
    <*> (optional $ strOption (long "dns" <> metavar "DNS"
                               <> help "metadata: DNS address"))
    <*> (switch $ long "nostart" <> help "define domain, but do not start it"))

parseRemove :: Parser Command
parseRemove = Remove <$> (RemoveOptions
    <$> (argument str (metavar "VM_NAME")))

parseOptions :: Parser Options
parseOptions = Options
    <$> subparser
        (command "init" (parseInit `withInfo` "Init qemu/KVM environment")
         <> command "create" (parseCreate `withInfo` "Create a new VM")
         <> command "rm" (parseRemove `withInfo`
                          "Destroy and undefine a VM. All data will be lost."))
