{-# LANGUAGE QuasiQuotes, OverloadedStrings #-}

module Domain where

import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import Text.Printf
import Text.Shakespeare.Text

import Control.Monad.State
import Data.Maybe (fromJust)

import Types
import Config
import RenderFile

makeDomain :: QvmDomain
makeDomain =
  QvmDomain {name = "unnamed"
            ,title = "no title"
            ,core = 2
            ,ramSizeMB = 512
            ,network = DefaultNAT
            ,disk =
               HTTPImage "http://dfs.game.yy.com:8080/download/cirros-0.3.4-x86_64-disk.img"
            ,ipAddress = Nothing
            ,gateway = Nothing
            ,dns = Nothing}

setName :: String -> State QvmDomain ()
setName n =
  do d <- get
     put d {name = n}

setCore :: Int -> State QvmDomain ()
setCore n =
  do d <- get
     put d {core = n}

setRAMMB :: Int -> State QvmDomain ()
setRAMMB ram =
  do d <- get
     put d {ramSizeMB = ram}

setRAMGB :: Int -> State QvmDomain ()
setRAMGB ram =
  do d <- get
     put d {ramSizeMB = ram * 1024}

setNetwork :: QvmNetwork -> State QvmDomain ()
setNetwork network =
  do d <- get
     put d {network = network}

setDisk :: QvmDisk -> State QvmDomain ()
setDisk disk =
  do d <- get
     put d {disk = disk}

renderDisk :: QvmDisk -> T.Text
renderDisk (LocalImage fn) =
  T.pack $
  printf "<disk type='file' device='disk'>\n\
         \  <driver name='qemu' type='qcow2' cache='writeback'/>\n\
         \  <source file='%s'/>\n\
         \  <target dev='vda' bus='virtio'/>\n\
         \</disk>"
         fn
renderDisk (HTTPImage _) =
  error "HTTPImage can not be rendered to domain xml."
renderDisk (SheepdogImage host port name) =
  T.pack $
  printf "<disk type='network' device='disk'>\n\
         \  <driver name='qemu' type='raw' cache='none'/>\n\
         \  <source protocol='sheepdog' name='%s'>\n\
         \    <host name='%s' port='%s'/>\n\
         \  </source>\n\
         \  <target dev='vda' bus='virtio'/>\n\
         \</disk>"
         name host port

renderNetwork :: QvmNetwork -> T.Text
renderNetwork DefaultNAT = renderNetwork $ DefinedNetwork "default"
renderNetwork (DefinedNetwork name) =
  T.pack $
  printf "<interface type='network'>\n\
      \  <source network='%s'/>\n\
      \  <model type='virtio'/>\n\
    \</interface>"
         name
renderNetwork (OvsBridge bridge) =
  T.pack $
  printf "<interface type='bridge'>\n\
        \  <forward mode='bridge'/>\n\
        \  <source bridge='%s'/>\n\
        \  <virtualport type='openvswitch'/>\n\
        \  <model type='virtio'/>\n\
        \</interface>"
         bridge
renderNetwork (OvsBridgeWithVLAN bridge vlan) =
  T.pack $
  printf "<interface type='bridge'>\n\
        \  <forward mode='bridge'/>\n\
        \  <source bridge='%s'/>\n\
        \  <virtualport type='openvswitch'/>\n\
        \  <vlan trunk='no'>\n\
        \    <tag id='%d'/>\n\
        \  </vlan>\n\
        \  <model type='virtio'/>\n\
        \</interface>"
         bridge vlan

-- | add N space in beginning of every line of given text. ignore empty lines.
shiftright :: Int -> T.Text -> T.Text
shiftright n text = T.unlines $ map (prependnspace n) (T.lines text) where
    prependnspace :: Int -> T.Text -> T.Text
    prependnspace n line = if T.null line
                           then line
                           else T.append (T.replicate n " ") line

renderXML :: QvmDomain -> T.Text
renderXML d = [stFile|src/domain-templates/domain.xml|]

printXML :: QvmDomain -> IO ()
printXML = TIO.putStrLn . renderXML

renderMetadata :: QvmDomain -> T.Text
renderMetadata d =
    case ipAddress d of
      Nothing -> [stFile|src/domain-templates/meta-data-no-ip|]
      _ -> [stFile|src/domain-templates/meta-data-ip|]

renderUserdata :: QvmDomain -> T.Text
renderUserdata d = [stFile|src/domain-templates/user-data|]

-- | incrementally setup a libvirt domain.
domainSetup :: State QvmDomain ()
domainSetup = do
  setName "vm1"
  setCore 1
  setRAMMB 512
  setCore 4
  setRAMGB 2
  setNetwork $ DefinedNetwork "ovs1"
  setDisk $
    LocalImage "/data/qvm/images/cirros-0.3.4-x86_64-disk.img"

-- | test whether domainSetup works, by printing out the result.
testDomain :: IO ()
testDomain =
  let d = execState domainSetup makeDomain
  in do print d
        putStrLn $ T.unpack $ renderXML d
