{-# LANGUAGE OverloadedStrings #-}

module DomainOperation where

import qualified Data.ByteString.Char8 as B
-- import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Lazy.Char8 as LB
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import qualified Crypto.Hash.SHA1 as SHA1
import Data.ByteString.Builder
import System.Directory
import System.FilePath.Posix
import System.Process
import System.Exit
import Control.Monad.Extra

import Types
import Domain
import Config
import DownloadFile

-- | infer filename suffix from given URL. suffix includes dot.
getFilenameSuffix :: String -> String
getFilenameSuffix = takeExtension

-- | run sha1 on given string, return same result as sha1sum command.
easySha1 :: String -> String
easySha1 s = LB.unpack $ toLazyByteString $ byteStringHex $ SHA1.hash $ B.pack s

-- | download a file from given url, save to localfile. return Right localfile
-- on success, return Left errmsg on failure. if localfile already exists,
-- just return it.
downloadFileMaybe :: String -> String -> IO (Either String String)
downloadFileMaybe url localfile =
  ifM (doesFileExist localfile)
    (return $ Right localfile) $
    do
      putStrLn $ "Downloading " ++ url
      downloadFile url localfile

-- | clone qcow2 disk, so VM will have its own boot disk.
-- this does a shallow clone. target disk is based on original disk.
cloneQcow2Disk :: String -> String -> IO ()
cloneQcow2Disk src dst = do
  putStrLn $ "Creating disk " ++ dst
  callProcess "sudo" (words "qemu-img create -f qcow2 -b" ++ [src, dst])

-- | clone localfile to get a root disk for given domain.
-- this does a shallow clone. target disk is based on original disk.
cloneDisk :: String -> QvmDomain -> IO QvmDomain
cloneDisk localfile d = do
  let suffix = getFilenameSuffix localfile
      vmfile = diskFile d suffix
  r <- doesFileExist vmfile
  if r then
    error ("Error: VM disk file " ++ vmfile ++ " exists. Exiting.")
  else do
    createDirectoryIfMissing True (takeDirectory vmfile)
    cloneQcow2Disk localfile vmfile
    return $ d {disk=LocalImage vmfile}

-- | clone given disk, return a LocalImage type of disk that can be used in
-- the domain.
getDisk :: QvmDomain -> IO QvmDomain
getDisk d =
    case disk d of
      LocalImage f -> do
                  ifM (doesFileExist f)
                    (cloneDisk f d)
                    (error $ "Error: local image does not exist: " ++ f)
      HTTPImage url -> do
                   let sha1 = easySha1 url
                       suffix = getFilenameSuffix url
                       originalname = takeFileName url
                       localfile = imageFile (sha1 ++ "-" ++ originalname)
                       vmfile = diskFile d suffix
                   createDirectoryIfMissing True (takeDirectory localfile)
                   r <- downloadFileMaybe url localfile
                   cloneDisk localfile d

prepareMetadata :: QvmDomain -> IO ()
prepareMetadata d = do
  createDirectoryIfMissing True (metadataDir d)
  TIO.writeFile (metadataFile d) (renderMetadata d)
  TIO.writeFile (userdataFile d) (renderUserdata d)
  createProcess $ (proc "genisoimage" (words "-input-charset utf-8 -output" ++ [metadataISOFile d] ++ words "-volid cidata -joliet -rock user-data meta-data")) {cwd=Just $ metadataDir d}
  return ()

-- | defines domain in libvirt. resolve any disk/network spec to the format
-- that libvirt can understand.
defineDomain1 :: QvmDomain -> IO QvmDomain
defineDomain1 d = do
    d <- getDisk d
    let xml = renderXML d
        domainfile = domainFile d
    ifM (doesFileExist domainfile)
      (error $ "Error: domain exists at " ++ domainfile) $
      do
        prepareMetadata d
        TIO.writeFile domainfile xml
        callProcess "sudo" ["virsh", "define", domainfile]
        return d

-- | run some pre-condition checks, then call defineDomain1 to do the hard
-- work.
defineDomain :: QvmDomain -> IO QvmDomain
defineDomain d =
  ifM (isDomainDefined $ name d)
    (error $ "Error: domain " ++ name d ++
             " is already defined. pick another name.") $
  ifM (doesFileExist $ diskFile d [])
    (error $ "Error: domain disk file exists: " ++ diskFile d []) $
  ifM (doesFileExist $ domainFile d)
    (error $ "Error: domain.xml file exists: " ++ domainFile d) $
  (defineDomain1 d)

startDomain :: QvmDomain -> IO ()
startDomain d =
    callProcess "sudo" ["virsh", "start", name d]

listDomains :: IO ()
listDomains = callCommand "ls -1 -d /data/qvm/instances/"

-- | return True if domain is defined in libvirt
isDomainDefined :: String -> IO Bool
isDomainDefined dname = do
  output <- readProcess "sudo" ["virsh", "list", "--name", "--all"] []
  return $ dname `elem` words output

-- | get domain's state
domainState :: String -> IO DomainState
domainState dname = do
  (exit, stdout, stderr) <- readProcessWithExitCode "sudo"
                            ["virsh", "domstate", dname] []
  case exit of
    ExitSuccess -> return $ case head (lines stdout) of
                     "running" -> Running
                     "shut off" -> Shutoff
                     "paused" -> Paused
                     otherwise -> Unknown
    ExitFailure _ -> return Unknown

-- | destroy, undefine domain, then remove all VM's files. This will delete VM
-- data.
removeDomain :: String -> IO ()
removeDomain dname = do
  whenM (fmap (==Running) (domainState dname)) $
    callProcess "sudo" ["virsh", "destroy", dname]
  whenM (isDomainDefined dname) $
    callProcess "sudo" ["virsh", "undefine", dname]
  removeDirectoryRecursive $ domainDir dname
